import json
import logging
import logging.config
import yaml
import os
import uuid
from datetime import datetime

import jwt
import tornado.ioloop
import tornado.web
import tornado.websocket
from jwt.algorithms import RSAAlgorithm
from tornado import iostream, gen
from tornado.log import app_log as log
from tornado.options import define
from tornado.web import HTTPError
import requests

import packer.jobs as jobs
from packer.file_handling import FSHandler
from packer.task_status import Status, TaskStatusAsync
from .config import tornado_config, app_config, logging_config, keycloak_config
from .redis_client import get_async_redis
from .tasks import app
import functools


def get_current_user(self):
    """ output of this is accessible in requests as self.current_user """
    token = get_request_token(self)
    try:
        if token is None or len(token) == 0:
            error_msg = 'No authorisation token found in the request'
            log.error(error_msg)
            raise HTTPError(401, 'Unauthorized.')
        decoded_token_header = jwt.get_unverified_header(token)
        token_kid = decoded_token_header.get('kid')
        algorithm, public_key = get_keycloak_public_key_and_algorithm(token_kid)
        user_token = jwt.decode(token, public_key, algorithms=algorithm, audience=keycloak_config.get("client_id"))
        subject = user_token.get('sub')
        log.info(f'Connected: {user_token.get("email")!r}, user id (sub): {subject!r}')
        return subject
    except Exception as e:
        raise HTTPError(401, f'Access unauthorized. {str(e)}')


def get_request_token(request_holder):
    token = request_holder.request.headers.get("Authorization")
    if token:
        return token.split('Bearer ')[-1]
    return None


@functools.lru_cache(maxsize=2)
def get_keycloak_public_key_and_algorithm(token_kid):
    handle = f'{keycloak_config.get("oidc_server_url")}/protocol/openid-connect/certs'
    log.info(f'Getting public key for the kid={token_kid} from the keycloak...')
    r = requests.get(handle)
    if r.status_code != 200:
        error = "Could not get certificates from Keycloak. " \
                "Reason: [{}]: {}".format(r.status_code, r.text)
        logging.error(error)
        raise ValueError(error)
    try:
        json_response = r.json()
    except Exception:
        error = "Could not retrieve the public key. " \
                "Got unexpected response: '{}'".format(r.text)
        logging.error(error)
        raise ValueError(error)
    try:
        matching_key = next((item for item in json_response.get('keys') if item['kid'] == token_kid), None)
        if matching_key is None:
            error = "No public key found for kid {}".format(token_kid)
            logging.error(error)
            raise ValueError(error)
        matching_key_json = json.dumps(matching_key)
        public_key = RSAAlgorithm.from_jwk(matching_key_json)
    except Exception as e:
        error = f'Invalid public key!. Reason: {e}'
        logging.error(error)
        raise ValueError(error)
    log.info(f'The public key for the kid={token_kid} has been fetched.')
    return matching_key.get('alg'), public_key


def setup_logging(default_level=logging.INFO):
    # Setup logging configuration
    path = logging_config.get('path', 'packer/logging.yaml')
    if os.path.exists(path):
        with open(path, 'rt') as f:
            log.info(f'Setting logging based on {path} configuration')
            config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)
    else:
        log.info('Logging configuration not found. Setting basic console log level.')
        logging.basicConfig(level=default_level)


class BaseHandler(tornado.web.RequestHandler):
    get_current_user = get_current_user

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", app_config.get("host"))
        self.set_header("Access-Control-Allow-Credentials", "true")
        self.set_header("Access-Control-Allow-Headers", "authorization, content-type")
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, OPTIONS')

    @property
    def user_jobs_key(self):
        return f'jobs:{self.current_user}'

    async def get_task_status(self, task_id):
        """
        Checks whether a task exists and current user has permissions to view.
        Will raise 404 if not, else will return the object to control it.

        :param task_id: uuid
        :return: TaskStatus object.
        """
        is_job = await self.application.redis.sismember(self.user_jobs_key, task_id)
        if not is_job:
            raise HTTPError(404, f'There is no task with id {task_id!r}.')
        else:
            return TaskStatusAsync(task_id, self.application.redis)

    async def options(self, *args):
        # no body
        self.set_status(200)
        self.finish()

    def write_error(self, status_code, **kwargs):
        self.set_status(status_code)
        self.finish({"error": str(kwargs['exc_info'][1])})


class JobListHandler(BaseHandler):
    """
    Provides a list of all jobs, current and past for current user.
    """

    async def get(self):
        log.info(f'Getting jobs for user: {self.current_user}')
        jobs_ = await self.application.redis.smembers(f'jobs:{self.current_user}')
        self.write(
            {'jobs': [await TaskStatusAsync(job, self.application.redis).get() for job in jobs_],
             'available_job_types': [job for job in jobs.registry.keys()]}
        )
        self.finish()


class CreateJobHandler(BaseHandler):
    """
    Start any available job type.
    """

    async def create(self, job_type: str, job_parameters: dict, **kwargs):

        log.info(f'New job request ({job_type}) for user: {self.current_user}')

        if kwargs:
            msg = f'Illegal arguments provided: {", ".join([k for k in kwargs.keys()])}.'
            log.info(msg)
            raise HTTPError(400, msg)

        if not isinstance(job_parameters, dict):
            msg = f'Unexpected argument, "job_parameters" should be in ' \
                  f'dict-like, but got {type(job_parameters)}'
            log.error(msg)
            raise HTTPError(400, msg)

        # Find the right job, and check parameters
        task = jobs.registry.get(job_type)

        if task is None:
            raise HTTPError(404, f'Job {job_type!r} not found.')
        log.info(f'Job {job_type!r} found.')

        task_id = str(uuid.uuid4())  # Job id used for tracking.
        await self.application.redis.sadd(self.user_jobs_key, task_id)

        task_status = await self.get_task_status(task_id)
        await task_status.create(
            job_type=job_type,
            job_parameters=job_parameters,
            status=Status.REGISTERED,
            user=self.current_user,
            created_at=datetime.utcnow().isoformat(sep='T', timespec='seconds') + 'Z'
        )

        try:
            task.apply_async(
                kwargs=job_parameters,
                task_id=task_id,
                headers=self.request.headers
            )
        except Exception as e:
            raise tornado.web.HTTPError(400, str(e))

        self.write(await task_status.get())
        self.finish()

    async def get(self):
        kwargs = self.request.arguments
        kwargs['job_type'] = self.get_argument('job_type')
        try:
            kwargs['job_parameters'] = json.loads(self.get_argument('job_parameters'))
        except json.JSONDecodeError:
            raise HTTPError(400, 'Expected valid JSON job parameters.')

        await self.create(**kwargs)

    async def post(self):
        try:
            kwargs = json.loads(self.request.body)
        except json.JSONDecodeError:
            raise HTTPError(400, 'Expected are valid JSON parameters.')

        await self.create(**kwargs)


class JobStatusHandler(BaseHandler):
    """
    Returns status object for single task.
    """

    async def get(self, task_id):
        task_status = await self.get_task_status(task_id)
        self.write(await task_status.get())
        self.finish()


class JobCancelHandler(BaseHandler):
    """
    Returns status object for single task.
    """

    async def get(self, task_id):
        task_status = await self.get_task_status(task_id)
        app.control.revoke(task_id, terminate=True, signal='SIGUSR1')
        logging.info(f'Cancel signal sent to worker for task: {task_id}')
        await task_status.update(
            status=Status.CANCELLED,
            message='Cancelled prior to execution.'
        )
        self.finish()


class DataHandler(BaseHandler):
    """
    Returns status object for single task.
    """

    async def get(self, task_id):
        task_status = await self.get_task_status(task_id)
        task_status = await task_status.get()

        # Should never happen, but let's check to be sure.
        if task_status['user'] != self.current_user:
            raise HTTPError(401, 'Unauthorized.')

        if task_status['status'] != Status.SUCCESS:
            raise HTTPError(403, f'Wrong task status ({task_status["status"]}), '
                                 f'has to be {Status.SUCCESS}.')

        file = FSHandler(task_id)

        if not file.exists():
            raise HTTPError(404, 'Resource not found. Contact administrator.')

        chunk_size = 1024 * 1024 * 1  # 1 MiB chunk size to read

        with file.byte_reader as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                try:
                    self.write(chunk)
                    await self.flush()

                except iostream.StreamClosedError:
                    break  # the client has closed the connection

                finally:
                    del chunk
                    # pause the coroutine so other handlers can run
                    await gen.sleep(0.00000001)  # 1 nanosecond

        self.finish()


class StatusWebSocket(tornado.websocket.WebSocketHandler):
    get_current_user = get_current_user

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sub = None

    def check_origin(self, origin):
        # FIXME need to limit connections from specific GB
        return True

    async def open(self):
        log.info("WebSocket opened")
        self.sub, = await self.application.redis.subscribe(f'channel:{self.current_user}')

        open_msg = f'Listening to channel: {self.sub.name.decode()!r}'
        log.info(open_msg)
        self.write_message(open_msg)

        async def async_reader(channel):
            while await channel.wait_message():
                msg = await channel.get()
                logging.info(msg)
                self.write_message(msg)

        await async_reader(self.sub)

    async def on_message(self, message):
        log.info(f"Message received: {message}")

    async def on_close(self):
        log.info("WebSocket closed by client.")
        await self.sub.unsubscribe()
        log.info("Unsubscribed from channel.")


class Application(tornado.web.Application):

    def __init__(self, *args, **kwargs):
        self.redis = None
        super().__init__(*args, **kwargs)

    def init_with_loop(self, loop):
        self.redis = get_async_redis(loop)


def make_web_app(port, tornado_options):
    """
    Create tornado app and loop.

    :return: (app, loop)
    """
    log.info('Creating web application.')
    web_app = Application([
        (r"/jobs", JobListHandler),
        (r"/jobs/create", CreateJobHandler),
        (r"/jobs/status/(.+)", JobStatusHandler),
        (r"/jobs/data/(.+)", DataHandler),
        (r"/jobs/cancel/(.+)", JobCancelHandler),
        (r"/jobs/subscribe", StatusWebSocket)
    ], **tornado_options)
    web_app.listen(port)
    loop = tornado.ioloop.IOLoop.current()
    web_app.init_with_loop(loop.asyncio_loop)
    return web_app, loop


def main():
    tornado.options.parse_command_line()
    port = tornado_config.get('port', 8888)
    setup_logging()
    web_app, loop = make_web_app(port, tornado_config)
    log.info(f'Starting at http://localhost:{web_app.settings.get("port")}')
    loop.start()



import json
import abc

from packer.redis_client import redis


class Status:
    REGISTERED = 'REGISTERED'
    CANCELLED = 'CANCELLED'
    FETCHING = 'FETCHING'
    RUNNING = 'RUNNING'
    SUCCESS = 'SUCCESS'
    FAILED = 'FAILED'


class TaskStatusABC(metaclass=abc.ABCMeta):

    def __init__(self, task_id):
        self.task_id = task_id
        self.key = f'job_status:{self.task_id}'

    @abc.abstractmethod
    def create(self, **kwargs):
        """ Store status for this task based on kwargs """

    @abc.abstractmethod
    def update(self, **kwargs):
        """ Update key-value pairs in kwargs """

    @abc.abstractmethod
    def get(self):
        """ Get the status as dictionary """


class TaskStatus(TaskStatusABC):

    def create(self, **kwargs):
        kwargs['task_id'] = self.task_id
        redis.set(self.key, json.dumps(kwargs))

    def update(self, **kwargs):
        obj = self.get()
        obj.update(**kwargs)
        self.create(**obj)

    def get(self):
        try:
            return json.loads(redis.get(self.key))
        except TypeError:
            return {}


class TaskStatusAsync(TaskStatusABC):

    def __init__(self, task_id, redis_loop):
        self.redis = redis_loop
        super().__init__(task_id)

    async def create(self, **kwargs):
        kwargs['task_id'] = self.task_id
        await self.redis.set(self.key, json.dumps(kwargs))

    async def update(self, **kwargs):
        obj = await self.get()
        obj.update(**kwargs)
        await self.create(**obj)

    async def get(self):
        status = await self.redis.get(self.key)
        return json.loads(status)




import pandas as pd


def filter_rows(df: pd.DataFrame, filter_row_df: pd.DataFrame) -> pd.DataFrame:
    """
    :param df: dataframe to apply fitering on
    :param filter_row_df: dataframe which index indicative on what rows should stay
    :return: dataframe as df but that has rows with index that appear in filter_row_df dataframe
    """
    df_indx_names = df.index.names
    filter_row_df_indx_names = filter_row_df.index.names
    common_indx_names = list(set(df_indx_names) & set(filter_row_df_indx_names))
    if not common_indx_names:
        raise ValueError('No index columns in common to filter rows.')
    df_common_indx = df.reset_index().set_index(common_indx_names).index
    filter_df_common_indx = filter_row_df.reset_index().set_index(common_indx_names).index
    filtered_df = df[df_common_indx.isin(filter_df_common_indx)]
    return filtered_df



import logging
import time

from packer.file_handling import FSHandler
from packer.task_status import Status
from ..tasks import BaseDataTask, app

logger = logging.getLogger(__name__)


def compute_addition(x: int, y: int):
    """
    This function contains the actual logic that is necessary to run the job.
    """
    return x + y


@app.task(bind=True, base=BaseDataTask)
def add(self: BaseDataTask, x: int, y: int, sleep: int = 10) -> None:
    """
    Example task that does basic additions of two integers.

    :param self: first argument is always required for bind to BaseDataTask.
    :param x: should be in job_parameters.
    :param y: should be in job_parameters.
    :param sleep: don't start for this many seconds (default 10).
    """

    if sleep:
        for i in range(sleep, 0, -1):
            msg = f'Task will be ready in {i} seconds'
            self.update_status(Status.RUNNING, msg)
            time.sleep(1)

    value = compute_addition(x, y)
    logger.info('Calculated value: {}'.format(value))

    with FSHandler(self.task_id).writer as writer:
        writer.write(value.to_bytes(2, byteorder='big'))

    logger.info(f'Stored to disk.')




import abc
import logging
import os

import json

from celery import Celery, Task
from celery.exceptions import SoftTimeLimitExceeded, Ignore

from packer.task_status import Status, TaskStatus
from .config import redis_config, task_config, celery_config, transmart_config
from .redis_client import redis

import requests

logger = logging.getLogger(__name__)


app = Celery('tasks', broker=redis_config['url'])
app.conf.update(**celery_config)
app.autodiscover_tasks(['packer.jobs'], 'jobs')

os.makedirs(task_config['data_dir'], exist_ok=True)


class BaseDataTask(Task, metaclass=abc.ABCMeta):

    def on_success(self, retval, task_id, args, kwargs):
        """Success handler.

        Run by the worker if the task executes successfully.

        Arguments:
            retval (Any): The return value of the task.
            task_id (str): Unique id of the executed task.
            args (Tuple): Original arguments for the executed task.
            kwargs (Dict): Original keyword arguments for the executed task.

        Returns:
            None: The return value of this handler is ignored.
        """
        self.update_status(status=Status.SUCCESS, message='Task finished successfully.')

    def on_retry(self, exc, task_id, args, kwargs, einfo):
        """Retry handler.

        This is run by the worker when the task is to be retried.

        Arguments:
            exc (Exception): The exception sent to :meth:`retry`.
            task_id (str): Unique id of the retried task.
            args (Tuple): Original arguments for the retried task.
            kwargs (Dict): Original keyword arguments for the retried task.
            einfo (~billiard.einfo.ExceptionInfo): Exception information.

        Returns:
            None: The return value of this handler is ignored.
        """

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        """Error handler.

        This is run by the worker when the task fails.

        Arguments:
            exc (Exception): The exception raised by the task.
            task_id (str): Unique id of the failed task.
            args (Tuple): Original arguments for the task that failed.
            kwargs (Dict): Original keyword arguments for the task that failed.
            einfo (~billiard.einfo.ExceptionInfo): Exception information.

        Returns:
            None: The return value of this handler is ignored.
        """
        if type(exc) == SoftTimeLimitExceeded:
            self.update_status(
                status=Status.CANCELLED,
                message='Task cancelled during execution or task passed time limit.'
            )
        else:
            self.update_status(
                status=Status.FAILED,
                message=f'Task failed with {exc.__class__.__name__}: {exc}'
            )

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        """Handler called after the task returns.

        Arguments:
            status (str): Current task state.
            retval (Any): Task return value/exception.
            task_id (str): Unique id of the task.
            args (Tuple): Original arguments for the task.
            kwargs (Dict): Original keyword arguments for the task.
            einfo (~billiard.einfo.ExceptionInfo): Exception information.

        Returns:
            None: The return value of this handler is ignored.
        """

    def __call__(self, *args, **kwargs):
        self.update_status(status=Status.RUNNING, message=f'Starting task.')
        super().__call__(*args, **kwargs)

    def get_data_dir(self, create=True):
        path = os.path.join(task_config['data_dir'], self.task_id)
        if create:
            os.makedirs(path, exist_ok=True)
        return path

    def open_writer(self, filename):
        return open(os.path.join(self.get_data_dir(), filename), 'w')

    @property
    def task_id(self):
        return self.request.id

    @property
    def task_status(self):
        return TaskStatus(self.task_id)

    @property
    def channel(self):
        obj = self.task_status.get()
        return f'channel:{obj.get("user")}'

    def update_status(self, status, message):
        """
        Send status update message through websocket, update job status in Redis.

        :param status: status code.
        :param message: message for client.
        """
        self.task_status.update(status=status, message=message)
        logger.info(f'Status update for {self.task_id}: {message} ({status})')
        redis.publish(
            self.channel,
            json.dumps({
                'task_id': self.task_id,
                'status': status,
                'message': message
            })
        )

    def observations_json(self, constraint):
        """
        :param self: Required for bind to BaseDataTask
        :param constraint: transmart API constraint to request
        :return: response body (json) of the observation call of transmart API
        """
        handle = f'{transmart_config.get("host")}/v2/observations'
        self.update_status(Status.FETCHING, f'Getting data from observations from {handle!r}')
        r = requests.post(url=handle,
                          json={'type': 'clinical', 'constraint': constraint},
                          headers={
                              'Authorization': self.request.get('Authorization')
                          })
        if r.status_code == 401:
            logger.error('Export failed. Unauthorized.')
            self.update_status(Status.FAILED, 'Unauthorized.')
            raise Ignore()
        if r.status_code != 200:
            logger.error('Export failed. Error occurred.')
            self.update_status(Status.FAILED, f'Connection error occurred when fetching {handle}. '
                                              f'Response status {r.status_code}')
            raise Ignore()
        return r.json()



import logging

import requests

from packer.export import save

try:
    from transmart.api.v2.data_structures import ObservationSet
except ImportError as e:
    logging.warning(f'Import errors for {__file__!r}: {str(e)}')

from packer.task_status import Status
from ..tasks import BaseDataTask, app

logger = logging.getLogger(__name__)


@app.task(bind=True, base=BaseDataTask)
def basic_export(self: BaseDataTask, constraint, **params):
    """
    Task that export transmart api client observation dataframe to tsv file.

    :param self: Required for bind to BaseDataTask
    :param constraint: should be in job_parameters.
    :param params: optional job parameters:
        - custom_name: name of the job and export file
    """
    obs_json = self.observations_json(constraint)
    self.update_status(Status.RUNNING, 'Observations gotten, transforming.')

    if 'custom_name' in params:
        custom_name = params['custom_name']
    else:
        logger.debug(f'No custom name supplied. Use task id as such {self.task_id}.')
        custom_name = self.task_id

    self.update_status(Status.RUNNING, 'Writing export to disk.')
    obs_df = ObservationSet(obs_json).dataframe
    save(obs_df, self.task_id, custom_name)



from .patient_diagnosis_biosource_biomaterial_export import patient_diagnosis_biosource_biomaterial_export
from .example import add
from .basic_export import basic_export

registry = {
    'add': add,
    'basic_export': basic_export,
    'patient_diagnosis_biosource_biomaterial_export': patient_diagnosis_biosource_biomaterial_export
}



import logging

from ..table_transformations.patient_diagnosis_biosource_biomaterial import from_obs_json_to_export_pdbb_df
from ..table_transformations.utils import filter_rows

from packer.task_status import Status
from ..tasks import BaseDataTask, app
from ..export import save

logger = logging.getLogger(__name__)


@app.task(bind=True, base=BaseDataTask)
def patient_diagnosis_biosource_biomaterial_export(self: BaseDataTask, constraint, **params):
    """
    Reformat hypercube data into a specific export file.
    Export table with the following IDs: Patient, Diagnosis, Biosource, Biomaterial

    :param self: Required for bind to BaseDataTask
    :param constraint: mandatory tranmsart api constraint to get data for.
    :param params: optional job parameters:
        - row_filter: constraint to filter rows
        - custom_name: name of the job and export file
    """
    obs_json = self.observations_json(constraint)
    self.update_status(Status.RUNNING, 'Observations gotten, transforming.')
    export_df = from_obs_json_to_export_pdbb_df(obs_json)
    if 'row_filter' in params:
        self.update_status(Status.RUNNING, 'Observations for the row filter gotten, transforming.')
        row_filter_constraint = params['row_filter']
        row_filter_obs_json = self.observations_json(row_filter_constraint)
        row_export_df = from_obs_json_to_export_pdbb_df(row_filter_obs_json)
        self.update_status(Status.RUNNING, 'Removing extra rows based on the row filter.')
        export_df = filter_rows(export_df, row_export_df)

    if 'custom_name' in params:
        custom_name = params['custom_name']
    else:
        logger.debug(f'No custom name supplied. Use task id as such {self.task_id}.')
        custom_name = self.task_id
    self.update_status(Status.RUNNING, 'Writing export to disk.')
    save(export_df, self.task_id, custom_name)




import re

import pandas as pd
import numpy as np
import logging

logger = logging.getLogger(__name__)
try:
    from transmart.api.v2.data_structures import ObservationSet
except ImportError as e:
    logging.warning(f'Import errors for {__file__!r}: {str(e)}')

DATE_FORMAT = '%Y-%m-%d'
ID_COLUMNS = ['Patient Id', 'Diagnosis Id', 'Biosource Id', 'Biomaterial Id']


def from_obs_json_to_export_pdbb_df(obs_json):
    """
    :param obs_json: json returned by transmart v2/observations call
    :return: data frame that has 4 (patient, diagnosis, biosource, biomaterial) index columns.
    The rest of columns represent concepts (aka variables)
    """

    df = ObservationSet(obs_json).dataframe
    concept_pat_to_name = _concept_path_to_name(df)
    df = from_obs_df_to_pdbb_df(df)
    df = df.rename(index=str, columns=concept_pat_to_name)
    df = format_columns(df)
    return df


def from_obs_df_to_pdbb_df(obs):
    if obs.empty:
        logger.warning('Retrieved hypercube is empty! Exporting empty result.')
        return obs
    # order rows by concept_paths:
    # 1)Patient -> 2)Diagnosis -> 3)Biosource -> 4)Biomaterial -> 5)Studies
    obs.rename(index=str, columns={'patient.trial': 'Patient Id'}, inplace=True)
    obs.sort_values(by=['concept.conceptPath'], inplace=True)
    concept_path_col = obs['concept.conceptPath']
    unq_concept_paths_ord = concept_path_col.unique().tolist()
    logger.info('Reformatting columns...')
    id_columns = _detect_index_columns(obs)
    obs = _reformat_columns(obs, id_columns)
    # transform concept rows to column headers
    obs_pivot = _concepts_row_to_columns(obs)
    # propagate data to lower levels and display only rows that represent the lowest level
    obs_pivot = _merge_redundant_rows(obs_pivot, id_columns)
    # fix columns order
    obs_pivot = obs_pivot[id_columns + unq_concept_paths_ord]
    # Replace NAs and NANs in index columns with empty string
    for id_column in id_columns:
        obs_pivot[id_column] = obs_pivot[id_column].fillna('')
    obs_pivot.set_index(id_columns, inplace=True)
    return obs_pivot


def _concept_path_to_name(df):
    return dict(zip(df['concept.conceptPath'], df['concept.name']))


def _detect_index_columns(df):
    lc_col_name_to_orig = {column.lower(): column for column in df.columns.values}
    return [lc_col_name_to_orig[id_column.lower()] for id_column in ID_COLUMNS if id_column.lower() in lc_col_name_to_orig]


def format_columns(df):
    """
    :param df: pandas dataframe with various data types of columns
    :return: modified data frame with all collumns converted to formatted string
    """
    result_df = pd.DataFrame()
    for col_num, col in enumerate(df.columns):
        # update datetime fields
        if re.match(r'.*\bdate\b.*', col, flags=re.IGNORECASE):
            result_df[col_num] = df.iloc[:, col_num].apply(_to_datetime)
        elif np.issubdtype(df.iloc[:, col_num].dtype, np.number):
            result_df[col_num] = df.iloc[:, col_num].apply(_num_to_str)
        else:
            result_df[col_num] = df.iloc[:, col_num]
    result_df = result_df.fillna('')
    result_df.columns=df.columns
    return result_df


def _num_to_str(x):
    if pd.isnull(x):
        return ''
    if isinstance(x, float) and x.is_integer():
        return str(int(x))
    return str(x)


def _merge_redundant_rows(data, id_columns):
    if data.empty:
        return
    # sort rows by identifying columns, merging of rows strongly depends on sorting
    rows = data.sort_values(id_columns, na_position='last').to_dict('records')
    result_rows = [rows[0]]
    for row in rows[1:]:
        row_copied = False
        for result_row in reversed(result_rows):
            if _is_ancestor_row(row, result_row, id_columns):
                _copy_missing_value_to_descendant_row(row, result_row, id_columns)
                row_copied = True
            else:
                break
        if not row_copied:
            result_rows.append(row)
    return pd.DataFrame(result_rows)


def _is_ancestor_row(ancestor_row_candidate, descendant_row_candidate, id_columns):
    for id_column in id_columns:
        if pd.isnull(ancestor_row_candidate[id_column]):
            break
        if ancestor_row_candidate[id_column] != descendant_row_candidate[id_column]:
            return False
    return True


def _copy_missing_value_to_descendant_row(ancestor_row, descendant_row, id_columns):
    for column, value in ancestor_row.items():
        if column in id_columns or pd.isnull(value):
            continue
        if column not in descendant_row or pd.isnull(descendant_row[column]):
            descendant_row[column] = ancestor_row[column]


def _to_datetime(date_str, string_format=DATE_FORMAT):
    if pd.notnull(date_str) and date_str is not None and date_str != '':
        try:
            return pd.to_datetime(date_str).strftime(string_format)
        except:
            return date_str
    else:
        return date_str


def _reformat_columns(obs, id_columns):
    # rename columns and set indexes
    obs.reset_index(inplace=True)
    headers = np.append(id_columns, 'concept.conceptPath')

    # prepare 'value' column
    if {'stringValue', 'numericValue'}.issubset(obs.columns):
        # merge stringValue and numericValue into value column
        obs['value'] = obs['stringValue'].fillna(obs['numericValue'])
        obs.drop(['stringValue', 'numericValue'], axis=1, inplace=True)
    elif 'stringValue' in obs:
        obs.rename(columns={"stringValue": "value"}, inplace=True)
    elif 'numericValue' in obs:
        obs.rename(columns={"numericValue": "value"}, inplace=True)
    else:
        obs['value'] = ""
    obs = obs.set_index(list(headers), append=True)[['value']]
    return obs


def _concepts_row_to_columns(obs):
    # use unstack to move the last level of the index to column names
    obs_pivot = obs.unstack(level=-1)
    # update column names by dropping value level
    obs_pivot.columns = obs_pivot.columns.droplevel(level=0)
    # fix indexes
    obs_pivot.reset_index(inplace=True)
    obs_pivot.drop(obs_pivot.columns[[0]], axis=1, inplace=True)
    return obs_pivot



import os

tornado_config = dict(
    cookie_secret="__TODO:_GENERATE_YOUR_OWN_RANDOM_VALUE_HERE__",
    debug=True,
    port=8999
)

keycloak_config = dict(
    oidc_server_url='{}/auth/realms/{}'.format(os.environ.get('KEYCLOAK_SERVER_URL'), os.environ.get('KEYCLOAK_REALM')),
    client_id=os.environ.get('KEYCLOAK_CLIENT_ID', 'transmart-client'),
)

transmart_config = dict(
    host=os.environ.get('TRANSMART_URL')
)

app_config = dict(
    host=os.environ.get('CLIENT_ORIGIN_URL', '*')
)

redis_config = dict(
    url=os.environ.get('REDIS_URL', 'redis://localhost:6379'),
)

task_config = dict(
    data_dir=os.environ.get('DATA_DIR', '/tmp/packer/')
)

celery_config = dict(
    task_serializer='json',
    accept_content=['json'],  # Ignore other content
    result_serializer='json',
    timezone='Europe/Amsterdam',
    enable_utc=True,
)

logging_config = dict(
    path=os.environ.get('LOG_CFG', 'packer/logging.yaml')
)



import aioredis
from redis import StrictRedis

from .config import redis_config

redis = StrictRedis.from_url(redis_config.get('url'), decode_responses=True)


def get_async_redis(loop):
    return loop.run_until_complete(
        aioredis.create_redis_pool(
            redis_config.get('url'),
            loop=loop,
            encoding='utf-8'
        )
    )



import csv
import pandas as pd
import logging

from packer.file_handling import FSHandler
from zipfile import ZipFile

logger = logging.getLogger(__name__)


def save(export_df: pd.DataFrame, task_id: str, file_name: str, sep: str = '\t'):
    """
    Writes dataframe including it's index columns to a file
    :param task_id: id of the task that indicates name of zip archive to store file to
    :param export_df: dataframe to write to file. Index columns are included.
    :param file_name: name of the file to export data to
    :param sep: separator in CSV file. Tab by default.
    """
    fs_handler = FSHandler(task_id)
    logger.info(f'Writing {fs_handler.path} file.')
    with fs_handler.writer as writer:
        with ZipFile(writer, 'w') as data_zip:
            data_zip.writestr(f'{file_name}.tsv',
                              export_df.reset_index().to_csv(encoding='utf-8', sep=sep, index=False,
                                                             quoting=csv.QUOTE_NONNUMERIC, quotechar='"'))
    logger.info(f'{fs_handler.path} file has been saved on disk.')



import abc
import os

from packer.config import task_config


class FileHandlerABC(metaclass=abc.ABCMeta):

    def __init__(self, task_id):
        self.task_id = task_id

    @abc.abstractmethod
    def exists(self):
        """ Returns True if the resource exists, else False """

    @property
    @abc.abstractmethod
    def reader(self):
        """ Return file-like object for reading """

    @property
    @abc.abstractmethod
    def byte_reader(self):
        """ Return file-like object for reading in byte mode """

    @property
    @abc.abstractmethod
    def writer(self):
        """ Return file-like object for writing """


class FSHandler(FileHandlerABC):
    """
    Provide method for reading and writing the output file of a task to filesystem.
    """

    @property
    def path(self):
        return os.path.join(task_config['data_dir'], self.task_id)

    def _handler(self, mode):
        return open(self.path, mode)

    def exists(self):
        return os.path.exists(self.path)

    @property
    def reader(self):
        return self._handler('r')

    @property
    def byte_reader(self):
        """ Read file as bytes. """
        return self._handler('rb')

    @property
    def writer(self):
        return self._handler('wb')
